//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file factor_point_feet_nomove.h
 *
 *  Created on: Feb 22, 2020
 *      \author: mfourmy
 */

#ifndef FACTOR_POINT_FEET_NOMOVE_H_
#define FACTOR_POINT_FEET_NOMOVE_H_

#include "core/math/rotations.h"
#include "core/factor/factor_autodiff.h"
#include "bodydynamics/sensor/sensor_point_feet_nomove.h"


namespace wolf
{

WOLF_PTR_TYPEDEFS(FactorPointFeetNomove);

class FactorPointFeetNomove : public FactorAutodiff<FactorPointFeetNomove, 
                                                    3,
                                                    3,
                                                    4,
                                                    3,
                                                    4
                                                    >
{
    public:
        FactorPointFeetNomove(const FeatureBasePtr&   _ftr_current_ptr,
                              const FrameBasePtr&     _frame_past_ptr,
                              const ProcessorBasePtr& _processor_ptr,
                              bool                    _apply_loss_function,
                              FactorStatus            _status = FAC_ACTIVE);

        ~FactorPointFeetNomove() override { /* nothing */ }

       /*
        Factor state blocks:
        _pb: W_p_WB -> base position in world frame
        _qb: W_q_B  -> base orientation in world frame
        _pb_past: W_p_WB -> base position in world frame, pastious KF
        _qb_past: W_q_B  -> base orientation in world frame, pastious KF
        */
        template<typename T>
        bool operator () (
            const T* const _pb,
            const T* const _qb,
            const T* const _pb_past,
            const T* const _qb_past,
            T* _res) const;

        Vector3d error() const;
        Vector3d res() const;
        double cost() const;

};

} /* namespace wolf */


namespace wolf {

FactorPointFeetNomove::FactorPointFeetNomove(
                            const FeatureBasePtr&   _ftr_current_ptr,
                            const FrameBasePtr&     _frame_past_ptr,
                            const ProcessorBasePtr& _processor_ptr,
                            bool                    _apply_loss_function,
                            FactorStatus            _status) :
    FactorAutodiff("FactorPointFeetNomove",
                   TOP_GEOM,
                   _ftr_current_ptr,
                    _frame_past_ptr,      // _frame_other_ptr
                    nullptr,              // _capture_other_ptr
                    nullptr,              // _feature_other_ptr
                    nullptr,              // _landmark_other_ptr
                    _processor_ptr,
                    _apply_loss_function,
                    _status,
                    _ftr_current_ptr->getFrame()->getP(),
                    _ftr_current_ptr->getFrame()->getO(),
                    _frame_past_ptr->getP(),
                    _frame_past_ptr->getO()
    )
{
}

Vector3d FactorPointFeetNomove::error() const{

    Vector3d pb = getFrame()->getP()->getState();
    Vector4d qb_vec = getFrame()->getO()->getState();
    Quaterniond qb(qb_vec);
    Vector3d pbm = getFrameOther()->getP()->getState();
    Vector4d qb_past_vec = getFrameOther()->getO()->getState();
    Quaterniond qbm(qb_past_vec);

    // Measurements retrieval
    Matrix<double,6,1> meas = getMeasurement();
    Vector3d bm_p_bml = meas.topRows(3);
    Vector3d b_p_bl      = meas.bottomRows(3);

    return (pbm + qbm*bm_p_bml) - (pb + qb*b_p_bl);

}

Vector3d FactorPointFeetNomove::res() const{
    return getMeasurementSquareRootInformationUpper() * error();
}


double FactorPointFeetNomove::cost() const{
    return res().squaredNorm();
}

template<typename T>
bool FactorPointFeetNomove::operator () (
                    const T* const _pb,
                    const T* const _qb,
                    const T* const _pb_past,
                    const T* const _qb_past,
                    T* _res) const
{
    Map<Matrix<T,3,1> > res(_res);

    // State variables instanciation
    Map<const Matrix<T,3,1> > pb(_pb);
    Map<const Quaternion<T> > qb(_qb);
    Map<const Matrix<T,3,1> > pbm(_pb_past);
    Map<const Quaternion<T> > qbm(_qb_past);

    // Measurements retrieval
    auto& meas = getMeasurement();
    Matrix<T,3,1> bm_p_bml = meas.segment(0, 3).cast<T>();       // Bminus_p_BminusL
    Matrix<T,4,1> bm_qvec_l = meas.segment(3, 4).cast<T>();      // Bminus_q_BminusL
    Matrix<T,3,1> b_p_bl = meas.segment(7, 3).cast<T>();         // B_p_BL
    Matrix<T,4,1> b_qvec_l = meas.segment(10, 4).cast<T>();      // B_q_BL
    Eigen::Quaternion<T> bm_q_l(bm_qvec_l);
    Eigen::Quaternion<T> b_q_l(b_qvec_l);

    // double radius = std::dynamic_pointer_cast<SensorPointFeetNomove>(getSensor())->getFootRadius();  // DOES NOT COMPILE
    double radius = 0.0;  // deactivate, was not doing so great anyway

    // If the radius of the foot is not negligeable, try to account for it with a more complex model
    Matrix<T,3,1> err;
    if (std::fabs(radius) < wolf::Constants::EPS){
        err = (pbm + qbm*bm_p_bml) - (pb + qb*b_p_bl);
    }
    else{
        /*
            n: frame with (xy) plane == world but orientation aligned in yaw with the robot base frame 
               sometimes called "Navigation frame" in some papers. 
            Simple roll without slipping model: we assume that most of actual movement
            of the frame attached to the center of the foot is due to rolling along y axis of the foot (for solo)
        */
        Quaternion<T> l_q_lm = (qbm*bm_q_l).inverse() * qb*b_q_l;  
        Matrix<T,3,1> l_aa_lm = log_q(l_q_lm);
        Matrix<T,3,1> n_p_lml = Matrix<T,3,1>::Zero(); 
        n_p_lml(0) = -l_aa_lm(1)*radius;

        // Factor: w_p_wl = w_p_wlm + w_R_n * n_p_lml 
        // By def, w_R_n is w_R_b with roll and pitch set to zero
        // Matrix<T,3,1> aa_bm = q2e(qbm); // DOES NOT WORK -> equivalent angle axis in this case
        Matrix<T,3,1> aa_bm = log_q(qbm);
        aa_bm.segment(0,2) = Matrix<T,2,1>::Zero();
        Quaternion<T> w_q_nm = exp_q(aa_bm); 

        err = (pbm + qbm*bm_p_bml) + w_q_nm*n_p_lml - (pb + qb*b_p_bl);
    }


    res = getMeasurementSquareRootInformationUpper() * err;

    return true;
}

} // namespace wolf

#endif /* FACTOR__POINT_FEET_NOMOVE_H_ */
