//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * imu_tools.h
 *
 *  Created on: Feb 17, 2020
 *      Author: mfourmy
 */

#ifndef FORCE_TORQUE_DELTA_TOOLS_H_
#define FORCE_TORQUE_DELTA_TOOLS_H_

#include "core/common/wolf.h"
#include "core/math/rotations.h"

/*
 * Most functions in this file are explained in the document:
 *
 *   TODO
 *
 * They relate manipulations of Delta motion magnitudes used for Force torque sensors pre-integration.
 *
 * The Delta is defined as
 *     Delta = [Dc, Dcd, DLc, Dq]
 * with
 *     Dc  : COM position delta
 *     Dcd : COM velocity delta
 *     DLc : Angular momentum delta
 *     Dq : quaternion delta
 *
 * They are listed below:
 *
 *   - identity: I = Delta at the origin, with Dp = [0,0,0]; Dq = [0,0,0,1], Dv = [0,0,0]
 *   - inverse: so that D (+) D.inv = I
 *   - compose: Dc = D1 (+) D2
 *   - between: Db = D2 (-) D1, so that D2 = D1 (+) Db
 *   - composeOverState: x2 = x1 (+) D
 *   - betweenStates: D = x2 (-) x1, so that x2 = x1 (+) D
 *   - log: got from Delta manifold to tangent space (equivalent to log() in rotations)
 *   - exp_FT: go from tangent space to delta manifold (equivalent to exp() in rotations)
 *   - plus: D2 = D1 (+) exp_imu(d)
 *   - diff: d = log_FT( D2 (-) D1 )
 *   - body2delta: construct a delta from body magnitudes
 */

namespace wolf 
{
namespace bodydynamics {
using namespace Eigen;

template<typename D1, typename D2, typename D3, typename D4>
inline void identity(MatrixBase<D1>& dc, MatrixBase<D2>& dcd, MatrixBase<D3>& dLc, QuaternionBase<D4>& dq)
{
    dc  = MatrixBase<D1>::Zero(3,1);
    dcd = MatrixBase<D3>::Zero(3,1);
    dLc = MatrixBase<D3>::Zero(3,1);
    dq  = QuaternionBase<D4>::Identity();
}

template<typename D1, typename D2, typename D3, typename D4>
inline void identity(MatrixBase<D1>& dc, MatrixBase<D2>& dcd, MatrixBase<D3>& dLc, MatrixBase<D4>& dq)
{
    typedef typename D1::Scalar T1;
    typedef typename D2::Scalar T2;
    typedef typename D3::Scalar T3;
    typedef typename D4::Scalar T4;
    dc  << T1(0), T1(0), T1(0);
    dcd << T2(0), T2(0), T2(0);
    dLc << T3(0), T3(0), T3(0);
    dq  << T4(0), T4(0), T4(0), T4(1);
}

template<typename T = double>
inline Matrix<T, 13, 1> identity()
{
    Matrix<T, 13, 1> ret;
    ret<< T(0), T(0), T(0),
          T(0), T(0), T(0),
          T(0), T(0), T(0),
          T(0), T(0), T(0), T(1);
    return ret;
}

template<typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8, class T>
inline void inverse(MatrixBase<D1>& dc, MatrixBase<D2>& dcd, MatrixBase<D3>& dLc, QuaternionBase<D4>& dq,
                    const T dt,
                    MatrixBase<D5>& idc, MatrixBase<D6>& idcd, MatrixBase<D7>& idLc, QuaternionBase<D8>& idq)
{
    MatrixSizeCheck<3, 1>::check(dc);
    MatrixSizeCheck<3, 1>::check(dcd);
    MatrixSizeCheck<3, 1>::check(dLc);
    MatrixSizeCheck<3, 1>::check(idc);
    MatrixSizeCheck<3, 1>::check(idcd);
    MatrixSizeCheck<3, 1>::check(idLc);

    idc =  - ( dq.conjugate() * (dc - dcd * typename D3::Scalar(dt)));
    idcd = - ( dq.conjugate() * dcd );
    idLc = - ( dq.conjugate() * dLc );
    idq =    ( dq.conjugate());
}

template<typename D1, typename D2, class T>
inline void inverse(const MatrixBase<D1>& d,
                    T dt,
                    MatrixBase<D2>& id)
{
    MatrixSizeCheck<13, 1>::check(d);
    MatrixSizeCheck<13, 1>::check(id);

    Map<const Matrix<typename D1::Scalar, 3, 1> >   dc   ( & d( 0 ) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dcd  ( & d( 3 ) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dLc  ( & d( 6 ) );
    Map<const Quaternion<typename D1::Scalar> >     dq   ( & d( 9 ) );
    Map<Matrix<typename D2::Scalar, 3, 1> >   idc   ( & id( 0 ) );
    Map<Matrix<typename D2::Scalar, 3, 1> >   idcd  ( & id( 3 ) );
    Map<Matrix<typename D2::Scalar, 3, 1> >   idLc  ( & id( 6 ) );
    Map<Quaternion<typename D1::Scalar> >     idq   ( & id( 9 ) );

    inverse(dc, dcd, dLc, dq, dt, idc, idcd, idLc, idq);
}

template<typename D, class T>
inline Matrix<typename D::Scalar, 13, 1> inverse(const MatrixBase<D>& d,
                                                 T dt)
{
    Matrix<typename D::Scalar, 13, 1> id;
    inverse(d, dt, id);
    return id;
}

template<typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8, typename D9, typename D10, typename D11, typename D12, class T>
inline void compose(const MatrixBase<D1>& dc1, const MatrixBase<D2>& dcd1, const MatrixBase<D3>& dLc1, const QuaternionBase<D4>& dq1,
                    const MatrixBase<D5>& dc2, const MatrixBase<D6>& dcd2, const MatrixBase<D7>& dLc2, const QuaternionBase<D8>& dq2,
                    const T dt,
                    MatrixBase<D9>& sum_c, MatrixBase<D10>& sum_cd, MatrixBase<D11>& sum_Lc, QuaternionBase<D12>& sum_q)
{
        MatrixSizeCheck<3, 1>::check(dc1);
        MatrixSizeCheck<3, 1>::check(dcd1);
        MatrixSizeCheck<3, 1>::check(dLc1);
        MatrixSizeCheck<3, 1>::check(dc2);
        MatrixSizeCheck<3, 1>::check(dcd2);
        MatrixSizeCheck<3, 1>::check(dLc2);
        MatrixSizeCheck<3, 1>::check(sum_c);
        MatrixSizeCheck<3, 1>::check(sum_cd);
        MatrixSizeCheck<3, 1>::check(sum_Lc);

        sum_c  = dc1  + dcd1*dt + dq1*dc2;
        sum_cd = dcd1 +           dq1*dcd2;
        sum_Lc = dLc1 +           dq1*dLc2;
        sum_q =                   dq1*dq2;
}

template<typename D1, typename D2, typename D3, class T>
inline void compose(const MatrixBase<D1>& d1,
                    const MatrixBase<D2>& d2,
                    T dt,
                    MatrixBase<D3>& sum)
{
    MatrixSizeCheck<13, 1>::check(d1);
    MatrixSizeCheck<13, 1>::check(d2);
    MatrixSizeCheck<13, 1>::check(sum);

    Map<const Matrix<typename D1::Scalar, 3, 1> >   dc1    ( & d1( 0 ) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dcd1   ( & d1( 3 ) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dLc1   ( & d1( 6 ) );
    Map<const Quaternion<typename D1::Scalar> >     dq1    ( & d1( 9 ) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dc2    ( & d2( 0 ) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dcd2   ( & d2( 3 ) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dLc2   ( & d2( 6 ) );
    Map<const Quaternion<typename D2::Scalar> >     dq2    ( & d2( 9 ) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         sum_c  ( & sum( 0 ) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         sum_cd  ( & sum( 3 ) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         sum_Lc  ( & sum( 6 ) );
    Map<Quaternion<typename D3::Scalar> >           sum_q  ( & sum( 9 ) );

    compose(dc1, dcd1, dLc1, dq1, dc2, dcd2, dLc2, dq2, dt, sum_c, sum_cd, sum_Lc, sum_q);
}

template<typename D1, typename D2, class T>
inline Matrix<typename D1::Scalar, 13, 1> compose(const MatrixBase<D1>& d1,
                                                  const MatrixBase<D2>& d2,
                                                  T dt)
{
    Matrix<typename D1::Scalar, 13, 1>  ret;
    compose(d1, d2, dt, ret);
    return ret;
}

template<typename D1, typename D2, typename D3, typename D4, typename D5, class T>
inline void compose(const MatrixBase<D1>& d1,
                    const MatrixBase<D2>& d2,
                    T dt,
                    MatrixBase<D3>& sum,
                    MatrixBase<D4>& J_sum_d1,
                    MatrixBase<D5>& J_sum_d2)
{
    MatrixSizeCheck<13, 1>::check(d1);
    MatrixSizeCheck<13, 1>::check(d2);
    MatrixSizeCheck<13, 1>::check(sum);
    MatrixSizeCheck< 12, 12>::check(J_sum_d1);
    MatrixSizeCheck< 12, 12>::check(J_sum_d2);

    // Some useful temporaries
    Map<const Quaternion<typename D1::Scalar> >     dq1    ( & d1( 9 ) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dc2    ( & d2( 0 ) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dcd2   ( & d2( 3 ) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dLc2   ( & d2( 6 ) );
    Map<const Quaternion<typename D2::Scalar> >     dq2    ( & d2( 9 ) );

    Matrix<typename D1::Scalar, 3, 3> dR1 = q2R(dq1);
    Matrix<typename D2::Scalar, 3, 3> dR2 = q2R(dq2);

    // // Jac wrt first delta
    J_sum_d1.setIdentity();                                     
    J_sum_d1.template block<3,3>(0,3) = Matrix3d::Identity() * dt;
    J_sum_d1.template block<3,3>(0,9) = - dR1 * skew(dc2);
    J_sum_d1.template block<3,3>(3,9) = - dR1 * skew(dcd2);
    J_sum_d1.template block<3,3>(6,9) = - dR1 * skew(dLc2);
    J_sum_d1.template block<3,3>(9,9) = dR2.transpose();

    // // // Jac wrt second delta
    J_sum_d2.setIdentity();
    J_sum_d2.template block<3,3>(0,0) = dR1;
    J_sum_d2.template block<3,3>(3,3) = dR1;
    J_sum_d2.template block<3,3>(6,6) = dR1;

    // compose deltas -- done here to avoid aliasing when calling with input 
    // `d1` and result `sum` referencing the same variable
    compose(d1, d2, dt, sum);
}

template<typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8, typename D9, typename D10, typename D11, typename D12, class T>
inline void between(const MatrixBase<D1>& dc1, const MatrixBase<D2>& dcd1, const MatrixBase<D3>& dLc1, const QuaternionBase<D4>& dq1,
                    const MatrixBase<D5>& dc2, const MatrixBase<D6>& dcd2, const MatrixBase<D7>& dLc2, const QuaternionBase<D8>& dq2,
                    const T dt,
                    MatrixBase<D9>& diff_c, MatrixBase<D10>& diff_cd, MatrixBase<D11>& diff_Lc, QuaternionBase<D12>& diff_q )
{
        MatrixSizeCheck<3, 1>::check(dc1);
        MatrixSizeCheck<3, 1>::check(dcd1);
        MatrixSizeCheck<3, 1>::check(dLc1);
        MatrixSizeCheck<3, 1>::check(dc2);
        MatrixSizeCheck<3, 1>::check(dcd2);
        MatrixSizeCheck<3, 1>::check(dLc2);
        MatrixSizeCheck<3, 1>::check(diff_c);
        MatrixSizeCheck<3, 1>::check(diff_cd);
        MatrixSizeCheck<3, 1>::check(diff_Lc);

        diff_c  = dq1.conjugate() * (dc2 - dc1 - dcd1*dt);
        diff_cd = dq1.conjugate() * (dcd2 - dcd1);
        diff_Lc = dq1.conjugate() * (dLc2 - dLc1);
        diff_q  = dq1.conjugate() * dq2;
}

template<typename D1, typename D2, typename D3, class T>
inline void between(const MatrixBase<D1>& d1,
                    const MatrixBase<D2>& d2,
                    T dt,
                    MatrixBase<D3>& d2_minus_d1)
{
    MatrixSizeCheck<13, 1>::check(d1);
    MatrixSizeCheck<13, 1>::check(d2);
    MatrixSizeCheck<13, 1>::check(d2_minus_d1);

    Map<const Matrix<typename D1::Scalar, 3, 1> >   dc1     ( & d1(0) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dcd1    ( & d1(3) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dLc1    ( & d1(6) );
    Map<const Quaternion<typename D1::Scalar> >     dq1     ( & d1(9) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dc2     ( & d2(0) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dcd2    ( & d2(3) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dLc2    ( & d2(6) );
    Map<const Quaternion<typename D2::Scalar> >     dq2     ( & d2(9) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   diff_c  ( & d2_minus_d1(0) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   diff_cd ( & d2_minus_d1(3) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   diff_Lc ( & d2_minus_d1(6) );
    Map<Quaternion<typename D3::Scalar> >     diff_q  ( & d2_minus_d1(9) );

    between(dc1, dcd1, dLc1, dq1, dc2, dcd2, dLc2, dq2, dt, diff_c, diff_cd, diff_Lc, diff_q);
}

template<typename D1, typename D2, class T>
inline Matrix<typename D1::Scalar, 13, 1> between(const MatrixBase<D1>& d1,
                                                  const MatrixBase<D2>& d2,
                                                  T dt)
{
    Matrix<typename D1::Scalar, 13, 1> diff;
    between(d1, d2, dt, diff);
    return diff;
}

template<typename D1, typename D2, typename D3, class T>
inline void composeOverState(const MatrixBase<D1>& x,
                             const MatrixBase<D2>& d,
                             T dt,
                             MatrixBase<D3>& x_plus_d)
{
    MatrixSizeCheck<13, 1>::check(x);
    MatrixSizeCheck<13, 1>::check(d);
    MatrixSizeCheck<13, 1>::check(x_plus_d);

    Map<const Matrix<typename D1::Scalar, 3, 1> >   c     ( & x(0) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   cd    ( & x(3) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   Lc    ( & x(6) );
    Map<const Quaternion<typename D1::Scalar> >     q     ( & x(9) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dc     ( & d(0) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dcd    ( & d(3) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dLc    ( & d(6) );
    Map<const Quaternion<typename D2::Scalar> >     dq     ( & d(9) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   c_plus_d  ( & x_plus_d(0) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   cd_plus_d ( & x_plus_d(3) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   Lc_plus_d ( & x_plus_d(6) );
    Map<Quaternion<typename D3::Scalar> >     q_plus_d  ( & x_plus_d(9) );

    c_plus_d = c + cd*dt + 0.5*gravity()*dt*dt + q*dc;
    cd_plus_d = cd + gravity()*dt + q*dcd;
    Lc_plus_d = Lc + q*dLc;
    q_plus_d = q*dq; // dq here to avoid possible aliasing between x and x_plus_d  --> WHAT?
}

template<typename D1, typename D2, class T>
inline Matrix<typename D1::Scalar, 13, 1> composeOverState(const MatrixBase<D1>& x,
                                                           const MatrixBase<D2>& d,
                                                           T dt)
{
    Matrix<typename D1::Scalar, 13, 1>  ret;
    composeOverState(x, d, dt, ret);
    return ret;
}

template<typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8, typename D9, typename D10, typename D11, typename D12, class T>
inline void betweenStates(const MatrixBase<D1>& c1, const MatrixBase<D2>& cd1, const MatrixBase<D3>& Lc1, const QuaternionBase<D4>& q1,
                          const MatrixBase<D5>& c2, const MatrixBase<D6>& cd2, const MatrixBase<D7>& Lc2, const QuaternionBase<D8>& q2,
                          const T dt,
                          MatrixBase<D9>& dc, MatrixBase<D10>& dcd, MatrixBase<D11>& dLc, QuaternionBase<D12>& dq)
{
        MatrixSizeCheck<3, 1>::check(c1);
        MatrixSizeCheck<3, 1>::check(cd1);
        MatrixSizeCheck<3, 1>::check(Lc1);
        MatrixSizeCheck<3, 1>::check(c2);
        MatrixSizeCheck<3, 1>::check(cd2);
        MatrixSizeCheck<3, 1>::check(Lc2);
        MatrixSizeCheck<3, 1>::check(dc);
        MatrixSizeCheck<3, 1>::check(dcd);
        MatrixSizeCheck<3, 1>::check(dLc);

        dc =  q1.conjugate() * ( c2 - c1 - cd1*dt - 0.5*gravity() *dt * dt );
        dcd = q1.conjugate() * ( cd2 - cd1        -   gravity()*dt );
        dLc = q1.conjugate() * ( Lc2 - Lc1);
        dq =  q1.conjugate() *   q2;
}

template<typename D1, typename D2, typename D3, class T>
inline void betweenStates(const MatrixBase<D1>& x1,
                          const MatrixBase<D2>& x2,
                          T dt,
                          MatrixBase<D3>& x2_minus_x1)
{
    MatrixSizeCheck<13, 1>::check(x1);
    MatrixSizeCheck<13, 1>::check(x2);
    MatrixSizeCheck<13, 1>::check(x2_minus_x1);

    Map<const Matrix<typename D1::Scalar, 3, 1> >   c1     ( & x1(0) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   cd1    ( & x1(3) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   Lc1    ( & x1(6) );
    Map<const Quaternion<typename D1::Scalar> >     q1     ( & x1(9) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   c2    ( & x2(0) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   cd2   ( & x2(3) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   Lc2   ( & x2(6) );
    Map<const Quaternion<typename D2::Scalar> >     q2    ( & x2(9) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   dc  ( & x2_minus_x1(0) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   dcd ( & x2_minus_x1(3) );
    Map<Matrix<typename D3::Scalar, 3, 1> >   dLc ( & x2_minus_x1(6) );
    Map<Quaternion<typename D3::Scalar> >     dq  ( & x2_minus_x1(9) );

    betweenStates(c1, cd1, Lc1, q1, c2, cd2, Lc2, q2, dt, dc, dcd, dLc, dq);
}

template<typename D1, typename D2, class T>
inline Matrix<typename D1::Scalar, 13, 1> betweenStates(const MatrixBase<D1>& x1,
                                                        const MatrixBase<D2>& x2,
                                                        T dt)
{
    Matrix<typename D1::Scalar, 13, 1> ret;
    betweenStates(x1, x2, dt, ret);
    return ret;
}

template<typename Derived>
Matrix<typename Derived::Scalar, 12, 1> log_FT(const MatrixBase<Derived>& delta_in)
{
    MatrixSizeCheck<13, 1>::check(delta_in);

    Matrix<typename Derived::Scalar, 12, 1> ret;

    Map<const Matrix<typename Derived::Scalar, 3, 1> >   dc_in  ( & delta_in(0) );
    Map<const Matrix<typename Derived::Scalar, 3, 1> >   dcd_in ( & delta_in(3) );
    Map<const Matrix<typename Derived::Scalar, 3, 1> >   dLc_in ( & delta_in(6) );
    Map<const Quaternion<typename Derived::Scalar> >     dq_in  ( & delta_in(9) );
    Map<Matrix<typename Derived::Scalar, 3, 1> >   dc_ret  ( & ret(0) );
    Map<Matrix<typename Derived::Scalar, 3, 1> >   dcd_ret ( & ret(3) );
    Map<Matrix<typename Derived::Scalar, 3, 1> >   dLc_ret ( & ret(6) );
    Map<Matrix<typename Derived::Scalar, 3, 1> >     do_ret  ( & ret(9) );

    dc_ret  = dc_in;
    dcd_ret = dcd_in;
    dLc_ret = dLc_in;
    do_ret  = log_q(dq_in);

    return ret;
}

template<typename Derived>
Matrix<typename Derived::Scalar, 13, 1> exp_FT(const MatrixBase<Derived>& d_alg_in)
{
    MatrixSizeCheck<12, 1>::check(d_alg_in);

    Matrix<typename Derived::Scalar, 13, 1> ret;

    Map<const Matrix<typename Derived::Scalar, 3, 1> >   dc_alg_in  ( & d_alg_in(0) );
    Map<const Matrix<typename Derived::Scalar, 3, 1> >   dcd_alg_in ( & d_alg_in(3) );
    Map<const Matrix<typename Derived::Scalar, 3, 1> >   dLc_alg_in ( & d_alg_in(6) );
    Map<const Matrix<typename Derived::Scalar, 3, 1> >   do_alg_in  ( & d_alg_in(9) );
    Map<Matrix<typename Derived::Scalar, 3, 1> >         dc_ret  ( & ret(0) );
    Map<Matrix<typename Derived::Scalar, 3, 1> >         dcd_ret ( & ret(3) );
    Map<Matrix<typename Derived::Scalar, 3, 1> >         dLc_ret ( & ret(6) );
    Map<Quaternion<typename Derived::Scalar> >           dq_ret  ( & ret(9) );

    dc_ret  = dc_alg_in;
    dcd_ret = dcd_alg_in;
    dLc_ret = dLc_alg_in;
    dq_ret  = exp_q(do_alg_in);

    return ret;
}

template<typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8, typename D9, typename D10, typename D11, typename D12>
inline void plus(const MatrixBase<D1>& dc1, const MatrixBase<D2>& dcd1, const MatrixBase<D3>& dLc1, const QuaternionBase<D4>& dq1, 
                 const MatrixBase<D5>& diff_c2, const MatrixBase<D6>& diff_cd2, const MatrixBase<D7>& diff_Lc2, const MatrixBase<D8>& diff_o2,
                 MatrixBase<D9>& plus_c, MatrixBase<D10>& plus_cd, MatrixBase<D11>& plus_Lc, QuaternionBase<D12>& plus_q)
{
        plus_c  = dc1  + diff_c2;
        plus_cd = dcd1 + diff_cd2;
        plus_Lc = dLc1 + diff_Lc2;
        plus_q  = dq1 * exp_q(diff_o2);
}


// Composite plus
template<typename D1, typename D2, typename D3>
inline void plus(const MatrixBase<D1>& d1,
                 const MatrixBase<D2>& diff2,
                 MatrixBase<D3>& d)
{
    Map<const Matrix<typename D1::Scalar, 3, 1> > dc1      ( & d1(0) );
    Map<const Matrix<typename D1::Scalar, 3, 1> > dcd1     ( & d1(3) );
    Map<const Matrix<typename D1::Scalar, 3, 1> > dLc1     ( & d1(6) );
    Map<const Quaternion<typename D1::Scalar> >   dq1      ( & d1(9) );
    Map<const Matrix<typename D2::Scalar, 3, 1> > diff_dc2   ( & diff2(0) );
    Map<const Matrix<typename D2::Scalar, 3, 1> > diff_dcd2  ( & diff2(3) );
    Map<const Matrix<typename D2::Scalar, 3, 1> > diff_dLc2  ( & diff2(6) );
    Map<const Matrix<typename D2::Scalar, 3, 1> > diff_do2   ( & diff2(9) );
    Map<Matrix<typename D3::Scalar, 3, 1> >       d_c  ( & d(0) );
    Map<Matrix<typename D3::Scalar, 3, 1> >       d_cd ( & d(3) );
    Map<Matrix<typename D3::Scalar, 3, 1> >       d_Lc ( & d(6) );
    Map<Quaternion<typename D3::Scalar> >         d_q  ( & d(9) );

    plus(dc1, dcd1, dLc1, dq1, 
         diff_dc2, diff_dcd2, diff_dLc2, diff_do2, 
         d_c, d_cd, d_Lc, d_q);
}

template<typename D1, typename D2>
inline Matrix<typename D1::Scalar, 13, 1> plus(const MatrixBase<D1>& d1,
                                               const MatrixBase<D2>& d2)
{
    Matrix<typename D1::Scalar, 13, 1> ret;
    plus(d1, d2, ret);
    return ret;
}

template<typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8, typename D9, typename D10, typename D11, typename D12>
inline void diff(const MatrixBase<D1>& dc1, const MatrixBase<D2>& dcd1, const MatrixBase<D3>& dLc1, const QuaternionBase<D4>& dq1,
                 const MatrixBase<D5>& dc2, const MatrixBase<D6>& dcd2, const MatrixBase<D7>& dLc2, const QuaternionBase<D8>& dq2,
                 MatrixBase<D9>& diff_c, MatrixBase<D10>& diff_cd, MatrixBase<D11>& diff_Lc, MatrixBase<D12>& diff_o)
{
        diff_c  = dc2 - dc1;
        diff_cd = dcd2 - dcd1;
        diff_Lc = dLc2 - dLc1;
        diff_o  = log_q(dq1.conjugate() * dq2);
}

template<typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8, typename D9, typename D10, typename D11, typename D12, typename D13, typename D14>
inline void diff(const MatrixBase<D1>& dc1, const MatrixBase<D2>& dcd1, const MatrixBase<D3>& dLc1, const QuaternionBase<D4>& dq1,
                 const MatrixBase<D5>& dc2, const MatrixBase<D6>& dcd2, const MatrixBase<D7>& dLc2, const QuaternionBase<D8>& dq2,
                 MatrixBase<D9>& diff_c, MatrixBase<D10>& diff_cd, MatrixBase<D11>& diff_Lc, MatrixBase<D12>& diff_o,
                 MatrixBase<D13>& J_do_dq1, MatrixBase<D14>& J_do_dq2)
{
    diff(dc1, dcd1, dLc1, dq1, dc2, dcd2, dLc2, dq2, diff_c, diff_cd, diff_Lc, diff_o);

    J_do_dq1 = - jac_SO3_left_inv(diff_o);
    J_do_dq2 =   jac_SO3_right_inv(diff_o);
}

template<typename D1, typename D2, typename D3>
inline void diff(const MatrixBase<D1>& d1,
                 const MatrixBase<D2>& d2,
                 MatrixBase<D3>& diff_d1_d2)
{
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dc1  ( & d1(0) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dcd1 ( & d1(3) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dLc1 ( & d1(6) );
    Map<const Quaternion<typename D1::Scalar> >     dq1  ( & d1(9) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dc2  ( & d2(0) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dcd2 ( & d2(3) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dLc2 ( & d2(6) );
    Map<const Quaternion<typename D1::Scalar> >     dq2  ( & d2(9) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         diff_c  ( & diff_d1_d2(0) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         diff_cd ( & diff_d1_d2(3) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         diff_Lc ( & diff_d1_d2(6) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         diff_o  ( & diff_d1_d2(9) );

    diff(dc1, dcd1, dLc1, dq1, dc2, dcd2, dLc2, dq2, diff_c, diff_cd, diff_Lc, diff_o);
}

template<typename D1, typename D2, typename D3, typename D4, typename D5>
inline void diff(const MatrixBase<D1>& d1,
                 const MatrixBase<D2>& d2,
                 MatrixBase<D3>& diff_d1_d2,
                 MatrixBase<D4>& J_diff_d1,
                 MatrixBase<D5>& J_diff_d2)
{
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dc1  ( & d1(0) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dcd1 ( & d1(3) );
    Map<const Matrix<typename D1::Scalar, 3, 1> >   dLc1 ( & d1(6) );
    Map<const Quaternion<typename D1::Scalar> >     dq1  ( & d1(9) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dc2  ( & d2(0) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dcd2 ( & d2(3) );
    Map<const Matrix<typename D2::Scalar, 3, 1> >   dLc2 ( & d2(6) );
    Map<const Quaternion<typename D1::Scalar> >     dq2  ( & d2(9) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         diff_c  ( & diff_d1_d2(0) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         diff_cd ( & diff_d1_d2(3) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         diff_Lc ( & diff_d1_d2(6) );
    Map<Matrix<typename D3::Scalar, 3, 1> >         diff_o  ( & diff_d1_d2(9) );

    Matrix<typename D4::Scalar, 3, 3> J_do_dq1, J_do_dq2;

    diff(dc1, dcd1, dLc1, dq1, dc2, dcd2, dLc2, dq2, diff_c, diff_cd, diff_Lc, diff_o, J_do_dq1, J_do_dq2);

    /* d = diff(d1, d2) is
     *   dc = dc2 - dc1
     *   dcd = dcd2 - dcd1
     *   dLc = dLc2 - dLc1
     *   do = Log(dq1.conj * dq2)
     *
     * With trivial Jacobians for dp and dv, and:
     *   J_do_dq1 = - J_l_inv(theta)
     *   J_do_dq2 =   J_r_inv(theta)
     */

    J_diff_d1 = - Matrix<typename D4::Scalar, 12, 12>::Identity();  // d(c2  - c1) / d(c1) = - Identity, for instance
    J_diff_d1.template block<3,3>(9,9) = J_do_dq1;                  // d(R1.tr*R2) / d(R1) = - J_l_inv(theta)
    J_diff_d2.setIdentity();                                        // d(c2  - c1) / d(c2) =   Identity, for instance
    J_diff_d2.template block<3,3>(9,9) = J_do_dq2;                  // d(R1.tr*R2) / d(R2) =   J_r_inv(theta)
}

template<typename D1, typename D2>
inline Matrix<typename D1::Scalar, 12, 1> diff(const MatrixBase<D1>& d1,
                                               const MatrixBase<D2>& d2)
{
    Matrix<typename D1::Scalar, 12, 1> ret;
    diff(d1, d2, ret);
    return ret;
}

template<typename D1, typename D2>
inline void body2delta(const MatrixBase<D1>& body,
                        const typename D1::Scalar& dt,
                        const typename D1::Scalar& mass,
                        int nbc,
                        int dimc,
                        MatrixBase<D2>& delta)
{
    typedef typename D1::Scalar T;

    delta.setZero();
    Map< Matrix<T, 3, 1>> dc ( & delta(0) );
    Map< Matrix<T, 3, 1>> dcd( & delta(3) );
    Map< Matrix<T, 3, 1>> dLc( & delta(6) );
    Map< Quaternion<T>>   dq ( & delta(9) );

    Vector3d pbc = body.segment(nbc*dimc, 3);
    Vector3d wb = body.segment(nbc*dimc + 3, 3);

    // general computation for any 3D or 6D contacts
    for (int i=0; i<nbc; i++){
        Vector3d pbli = body.segment(nbc*dimc + 6 + i*3, 3);
        Vector4d qbli_v = body.segment(nbc*dimc + 6 + nbc*3 + i*4, 4); 
        Quaterniond qbli(qbli_v);        
        Vector3d fi = body.segment(i*3, 3);
        
        dc += qbli*fi;
        dcd += qbli*fi;
        dLc += (pbli - pbc).cross(qbli * fi);
        if (dimc == 6){
            Vector3d taui = body.segment(3*nbc + i*3, 3);
            dLc += qbli*taui;
        }
    }

    // formulas for 2 6D contacts:
    // dc = 0.5 * (qbl1*f1 + qbl2*f2) * dt * dt / mass;
    // dcd =      (qbl1*f1 + qbl2*f2) * dt     / mass;
    // dLc = (qbl1*tau1 + (pbl1-pbc).cross(qbl1*f1) 
    //      + qbl2*tau2 + (pbl2-pbc).cross(qbl2*f2)) * dt;

    dc  *= (0.5*dt*dt/mass);
    dcd *= (dt/mass);
    dLc *= dt;
    dq = exp_q(wb * dt);
}

template<typename D1>
inline Matrix<typename D1::Scalar, 13, 1> body2delta(const MatrixBase<D1>& body,
                                                     const typename D1::Scalar& dt,
                                                     const typename D1::Scalar& mass,
                                                     int nbc,
                                                     int dimc)
{
    typedef typename D1::Scalar T;
    Matrix<T, 13, 1> delta;
    body2delta(body, dt, mass, nbc, dimc, delta);

    return delta;
}

template<typename D1, typename D2, typename D3>
inline void body2delta(const MatrixBase<D1>& body,
                       const typename D1::Scalar& dt,
                       const typename D1::Scalar& mass,
                       int nbc,
                       int dimc,
                       MatrixBase<D2>& delta,
                       MatrixBase<D3>& jac_delta_body)
{
    Vector3d pbc = body.segment(nbc*dimc, 3);
    Vector3d wb = body.segment(nbc*dimc + 3, 3);

    jac_delta_body.setZero();
    for (int i=0; i<nbc; i++){
        Vector3d pbli = body.segment(nbc*dimc + 6 + i*3, 3);
        Vector4d qbli_v = body.segment(nbc*dimc + 6 + nbc*3 + i*4, 4); 
        Quaterniond qbli(qbli_v);
        Matrix3d bRli = q2R(qbli);
        Vector3d fi = body.segment(i*3, 3);

        jac_delta_body.template block<3,3>(0,i*3) = 0.5 * bRli * dt * dt / mass;
        jac_delta_body.template block<3,3>(3,i*3) = bRli * dt / mass;
        jac_delta_body.template block<3,3>(6,i*3) = skew(pbli - pbc) * bRli * dt;
        if (dimc == 6){
            jac_delta_body.template block<3,3>(6,nbc*3 + i*3) = bRli * dt;
        }
        jac_delta_body.template block<3,3>(6,nbc*dimc) +=  skew(bRli*fi) * dt;
    }
    jac_delta_body.template block<3,3>(9,nbc*dimc+3) =  dt * jac_SO3_right(wb * dt);

    // formulas for 2 6D contacts:
    // jac_delta_body.template block<3,3>(0,0) = 0.5 * bRl1 * dt * dt / mass;
    // jac_delta_body.template block<3,3>(0,3) = 0.5 * bRl2 * dt * dt / mass;
    // jac_delta_body.template block<3,3>(3,0) =       bRl1 * dt / mass;
    // jac_delta_body.template block<3,3>(3,3) =       bRl2 * dt / mass;
    // jac_delta_body.template block<3,3>(6,0) =       skew(pbl1 - pbc) * bRl1 * dt;
    // jac_delta_body.template block<3,3>(6,3) =       skew(pbl2 - pbc) * bRl2 * dt;
    // jac_delta_body.template block<3,3>(6,6) =       bRl1 * dt;
    // jac_delta_body.template block<3,3>(6,9) =       bRl2 * dt;
    // jac_delta_body.template block<3,3>(6,12) =      (skew(bRl1*f1) + skew(bRl2*f2)) * dt;
    // jac_delta_body.template block<3,3>(9,15) =   dt * jac_SO3_right(wb * dt);

    // TODO: kinematics uncertainties (bRl1, bRl2, bpl1, bpl2) are not implemented currently

    body2delta(body, dt, mass, nbc, dimc, delta);
}

template<typename D1, typename D2, typename D3>
inline void debiasData(const MatrixBase<D1>& _data, const MatrixBase<D2>& _bias, int nbc, int dimc, MatrixBase<D3>& _body)
{
    MatrixSizeCheck<6 , 1>::check(_bias);
    _body = _data;
    _body.template block<6,1>(nbc*dimc,0) = _data.template block<6,1>(nbc*dimc,0) - _bias;
}

template<typename D1, typename D2, typename D3, typename D4>
inline void debiasData(const MatrixBase<D1>& _data, const MatrixBase<D2>& _bias, int nbc, int dimc, MatrixBase<D3>& _body, MatrixBase<D4>& _J_body_bias)
{
    debiasData(_data, _bias, nbc, dimc, _body);
    // MatrixSizeCheck<30, 6>::check(_J_body_bias);  // 30 because the 2 last body quantities are quaternions
    _J_body_bias.setZero();
    _J_body_bias.template block<3,3>(nbc*dimc, 0) = -Matrix<typename D1::Scalar, 3, 3>::Identity();
    _J_body_bias.template block<3,3>(nbc*dimc+3, 3) = -Matrix<typename D1::Scalar, 3, 3>::Identity();
}



} // namespace bodydynamics
} // namespace wolf

#endif /* FORCE_TORQUE_DELTA_TOOLS_H_ */
