//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
// wolf
#include "bodydynamics/math/force_torque_delta_tools.h"
#include "bodydynamics/capture/capture_force_torque.h"
#include "bodydynamics/feature/feature_force_torque.h"
#include "bodydynamics/processor/processor_force_torque.h"
#include "bodydynamics/factor/factor_force_torque.h"

namespace wolf
{

int compute_data_size(int nbc, int dimc)
{
    // compute the size of the data/body vectors used by processorMotion API depending
    // on the number of contacts (nbc) and contact dimension (dimc)
    // result: forces (+torques for 6D contacts) + pbc + wb + contact/base positions + contact/base orientations
    return nbc * dimc + 3 + 3 + nbc * 3 + nbc * 4;
}

using namespace Eigen;

ProcessorForceTorque::ProcessorForceTorque(ParamsProcessorForceTorquePtr _params_motion_force_torque)
    : ProcessorMotion("ProcessorForceTorque",
                      "CDLO",
                      3,
                      13,
                      13,
                      12,
                      compute_data_size(_params_motion_force_torque->nbc, _params_motion_force_torque->dimc),
                      6,
                      _params_motion_force_torque),
      params_motion_force_torque_(std::make_shared<ParamsProcessorForceTorque>(*_params_motion_force_torque)),
      nbc_(_params_motion_force_torque->nbc),
      dimc_(_params_motion_force_torque->dimc)

{
    //
}

ProcessorForceTorque::~ProcessorForceTorque()
{
    //
}

bool ProcessorForceTorque::voteForKeyFrame() const
{
    // time span
    if (getBuffer().back().ts_ - getBuffer().front().ts_ > params_motion_force_torque_->max_time_span)
    {
        WOLF_DEBUG("PM: vote: time span");
        return true;
    }
    // buffer length
    if (getBuffer().size() > params_motion_force_torque_->max_buff_length)
    {
        WOLF_DEBUG("PM: vote: buffer length");
        return true;
    }

    // Some other measure of movement?

    // WOLF_DEBUG( "PM: do not vote" );
    return false;
}

CaptureMotionPtr ProcessorForceTorque::emplaceCapture(const FrameBasePtr&   _frame_own,
                                                      const SensorBasePtr&  _sensor,
                                                      const TimeStamp&      _ts,
                                                      const VectorXd&       _data,
                                                      const MatrixXd&       _data_cov,
                                                      const VectorXd&       _calib,
                                                      const VectorXd&       _calib_preint,
                                                      const CaptureBasePtr& _capture_origin)
{
    // Here we have to retrieve the origin capture no
    // !! PROBLEM:
    // when doing setOrigin, emplaceCapture is called 2 times
    // - first on the first KF (usually created by setPrior), this one contains the biases
    // - secondly on the inner frame (last) which does not contains other captures
    auto capture_ikin   = _frame_own->getCaptureOf(sensor_ikin_);
    auto capture_angvel = _frame_own->getCaptureOf(sensor_angvel_);
    if ((capture_ikin == nullptr) || (capture_angvel == nullptr))
    {
        // We have to retrieve the bias from a former Keyframe: origin
        capture_ikin   = origin_ptr_->getFrame()->getCaptureOf(sensor_ikin_);
        capture_angvel = origin_ptr_->getFrame()->getCaptureOf(sensor_angvel_);
    }
    auto cap =
        CaptureBase::emplace<CaptureForceTorque>(_frame_own,
                                                 _ts,
                                                 _sensor,
                                                 std::static_pointer_cast<CaptureInertialKinematics>(capture_ikin),
                                                 std::static_pointer_cast<CaptureMotion>(capture_angvel),
                                                 _data,
                                                 _data_cov,
                                                 _capture_origin);

    // Cannot be recovered from the _calib and _calib_preint which are initialized using calib_size_
    // which is zero in our case since the calibration stateblocks are not actually in the FTPreint sensor/captures

    auto     cap_ft = std::static_pointer_cast<CaptureForceTorque>(cap);
    Vector6d calib  = getCalibration(cap_ft);
    setCalibration(cap_ft, calib);
    cap_ft->setCalibrationPreint(calib);

    return cap;
}

// FeatureBasePtr ProcessorForceTorque::emplaceFeature(CaptureMotionPtr _capture_motion)
// {
//     // Retrieving the origin capture and getting the bias from here should be done here?
//     auto feature = FeatureBase::emplace<FeatureForceTorque>(
//         _capture_motion,
//         _capture_motion->getBuffer().back().delta_integr_,
//         _capture_motion->getBuffer().back().delta_integr_cov_ + unmeasured_perturbation_cov_,
//         _capture_motion->getCalibrationPreint(),
//         _capture_motion->getBuffer().back().jacobian_calib_);
//     return feature;
// }

Eigen::VectorXd ProcessorForceTorque::correctDelta(const Eigen::VectorXd& delta_preint,
                                                   const Eigen::VectorXd& delta_step) const
{
    return bodydynamics::plus(delta_preint, delta_step);
}

VectorXd ProcessorForceTorque::getCalibration(const CaptureBaseConstPtr _capture) const
{
    VectorXd bias_vec(6);

    if (_capture)  // access from capture is quicker
    {
        auto cap_ft(std::static_pointer_cast<const CaptureForceTorque>(_capture));

        // get calib part from Ikin capture
        bias_vec.segment<3>(0) = cap_ft->getIkinCaptureOther()->getSensorIntrinsic()->getState();

        // get calib part from IMU capture
        bias_vec.segment<3>(3) =
            cap_ft->getGyroCaptureOther()
                ->getSensorIntrinsic()
                ->getState()
                .tail<3>();  // TODO: only valib for IMU bias sb or any other of the structure [XXXX, gyro_bias]
    }
    else  // access from sensor is slower
    {
        // get calib part from Ikin capture
        bias_vec.segment<3>(0) = sensor_ikin_->getIntrinsic()->getState();

        // get calib part from IMU capture
        bias_vec.segment<3>(3) =
            sensor_angvel_->getIntrinsic()
                ->getState()
                .tail<3>();  // TODO: only valib for IMU bias sb or any other of the structure [XXXX, gyro_bias]
    }
    return bias_vec;
}

void ProcessorForceTorque::setCalibration(const CaptureBasePtr _capture, const VectorXd& _calibration)
{
    CaptureForceTorquePtr cap_ft(std::static_pointer_cast<CaptureForceTorque>(_capture));

    // set calib part in Ikin capture
    cap_ft->getIkinCaptureOther()->getSensorIntrinsic()->setState(_calibration.head<3>());

    // set calib part in IMU capture
    Vector6d calib_imu  = cap_ft->getGyroCaptureOther()->getSensorIntrinsic()->getState();
    calib_imu.tail<3>() = _calibration.tail<3>();

    cap_ft->getGyroCaptureOther()->getSensorIntrinsic()->setState(calib_imu);
}

void ProcessorForceTorque::emplaceFeaturesAndFactors(CaptureBasePtr _capture_origin, CaptureMotionPtr _capture_own)
{
    // Retrieving the origin capture and getting the bias from here should be done here?
    auto feature = FeatureBase::emplace<FeatureForceTorque>(
        _capture_own,
        _capture_own->getBuffer().back().delta_integr_,
        _capture_own->getBuffer().back().delta_integr_cov_ + unmeasured_perturbation_cov_,
        _capture_own->getCalibrationPreint(),
        _capture_own->getBuffer().back().jacobian_calib_);

    CaptureForceTorquePtr cap_ft_origin = std::static_pointer_cast<CaptureForceTorque>(_capture_origin);
    FeatureForceTorquePtr ftr_ft        = std::static_pointer_cast<FeatureForceTorque>(feature);
    CaptureForceTorquePtr cap_ft_new    = std::static_pointer_cast<CaptureForceTorque>(ftr_ft->getCapture());

    FactorBase::emplace<FactorForceTorque>(ftr_ft,
                                           ftr_ft,
                                           cap_ft_origin,
                                           shared_from_this(),
                                           cap_ft_origin->getIkinCaptureOther(),  // to retrieve sb_bp1
                                           cap_ft_origin->getGyroCaptureOther(),  // to retrieve sb_w1
                                           false);
}

void ProcessorForceTorque::computeCurrentDelta(const Eigen::VectorXd& _data,
                                               const Eigen::MatrixXd& _data_cov,
                                               const Eigen::VectorXd& _calib,
                                               const double           _dt,
                                               Eigen::VectorXd&       _delta,
                                               Eigen::MatrixXd&       _delta_cov,
                                               Eigen::MatrixXd&       _jac_delta_calib) const
{
    assert(_data.size() == data_size_ && "Wrong data size!");

    // create delta
    MatrixXd jac_body_bias(data_size_ - nbc_, 6);
    VectorXd body(data_size_);
    bodydynamics::debiasData(_data, _calib, nbc_, dimc_, body, jac_body_bias);

    MatrixXd jac_delta_body(12, data_size_ - nbc_);
    bodydynamics::body2delta(body,
                             _dt,
                             std::static_pointer_cast<const SensorForceTorque>(getSensor())->getMass(),
                             nbc_,
                             dimc_,
                             _delta,
                             jac_delta_body);  // Jacobians tested in bodydynamics_tools

    // [f], [tau], pbc, wb
    MatrixXd jac_delta_body_reduced = jac_delta_body.leftCols(nbc_ * dimc_ + 6);

    // compute delta_cov (using uncertainty on f,tau,pbc)
    _delta_cov = jac_delta_body_reduced * _data_cov *
                 jac_delta_body_reduced.transpose();  // trivially:  jac_body_delta = Identity
    // _delta_cov = jac_delta_body * _data_cov * jac_delta_body.transpose();  // trivially:  jac_body_delta = Identity

    // compute jacobian_calib
    _jac_delta_calib = jac_delta_body * jac_body_bias;
}

void ProcessorForceTorque::deltaPlusDelta(const Eigen::VectorXd& _delta_preint,
                                          const Eigen::VectorXd& _delta,
                                          const double           _dt,
                                          Eigen::VectorXd&       _delta_preint_plus_delta) const
{
    _delta_preint_plus_delta = bodydynamics::compose(_delta_preint, _delta, _dt);
}

void ProcessorForceTorque::statePlusDelta(const VectorComposite& _x,
                                          const Eigen::VectorXd& _delta,
                                          const double           _dt,
                                          VectorComposite&       _x_plus_delta) const
{
    assert(_delta.size() == 13 && "Wrong _delta vector size");
    assert(_dt >= 0 && "Time interval _dt is negative!");

    // verbose way : create Eigen vectors, then compute, then convert back to Composite

    auto x = _x.vector("CDLO");

    VectorXd x_plus_delta = bodydynamics::composeOverState(x, _delta, _dt);

    _x_plus_delta = VectorComposite(x_plus_delta, "CDLO", {3, 3, 3, 4});
}

void ProcessorForceTorque::deltaPlusDelta(const Eigen::VectorXd& _delta_preint,
                                          const Eigen::VectorXd& _delta,
                                          const double           _dt,
                                          Eigen::VectorXd&       _delta_preint_plus_delta,
                                          Eigen::MatrixXd&       _jacobian_delta_preint,
                                          Eigen::MatrixXd&       _jacobian_delta) const
{
    bodydynamics::compose(_delta_preint,
                          _delta,
                          _dt,
                          _delta_preint_plus_delta,
                          _jacobian_delta_preint,
                          _jacobian_delta);  // jacobians tested in bodydynamics_tools
}

}  // namespace wolf

// Register in the FactorySensor
#include "core/processor/factory_processor.h"

namespace wolf
{
WOLF_REGISTER_PROCESSOR(ProcessorForceTorque)
}
