//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

/**
 * \file gtest_factor_inertial_kinematics.cpp
 *
 *  Created on: March 9, 2020
 *      \author: Mederic Fourmy
 */

// debug
#include <iostream>
#include <fstream>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>

// WOLF
#include <core/problem/problem.h>
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/math/rotations.h>
#include <core/capture/capture_base.h>
#include <core/capture/capture_pose.h>
#include <core/feature/feature_base.h>
#include <core/factor/factor_block_absolute.h>

// Imu
#include "imu/internal/config.h"
#include "imu/capture/capture_imu.h"
#include "imu/processor/processor_imu.h"
#include "imu/sensor/sensor_imu.h"

// BODYDYNAMICS
#include "bodydynamics/internal/config.h"
#include "bodydynamics/sensor/sensor_inertial_kinematics.h"
#include "bodydynamics/processor/processor_inertial_kinematics.h"
#include "bodydynamics/capture/capture_inertial_kinematics.h"

#include <Eigen/Eigenvalues>

using namespace wolf;
using namespace Eigen;
using namespace std;

const Eigen::Vector3d ZERO3 = Eigen::Vector3d::Zero();
const Eigen::Vector6d ZERO6 = Eigen::Vector6d::Zero();


class FactorInertialKinematics_2KF : public testing::Test
{
    public:
        wolf::TimeStamp t_;
        ProblemPtr problem_;
        SensorImuPtr sen_imu_;
        SolverCeresPtr solver_;
        SensorInertialKinematicsPtr sen_ikin_;
        VectorXd x_origin_;
        MatrixXd P_origin_;
        FrameBasePtr KF0_;

    void SetUp() override
    {
        std::string bodydynamics_root_dir = _WOLF_BODYDYNAMICS_ROOT_DIR;

        problem_ = Problem::create("POVCDL", 3);

        // CERES WRAPPER
        solver_ = std::make_shared<SolverCeres>(problem_);
        // solver_->getSolverOptions().minimizer_type = ceres::TRUST_REGION; //ceres::TRUST_REGION;ceres::LINE_SEARCH
        // solver_->getSolverOptions().max_line_search_step_contraction = 1e-3;
        // solver_->getSolverOptions().max_num_iterations = 1e4;

        //===================================================== INITIALIZATION
        t_.set(0.0);

        // SENSOR + PROCESSOR INERTIAL KINEMATICS
        ParamsSensorInertialKinematicsPtr intr_ik = std::make_shared<ParamsSensorInertialKinematics>();
        intr_ik->std_pbc = 0.1;
        intr_ik->std_vbc = 0.1;
        VectorXd extr;  // default size for dynamic vector is 0-0 -> what we need
        SensorBasePtr sen_ikin_base = problem_->installSensor("SensorInertialKinematics", "SenIK", extr, intr_ik);
        // SensorBasePtr sen_ikin_base = problem_->installSensor("SensorInertialKinematics", "SenIK", extr,  bodydynamics_root_dir + "/test/sensor_inertial_kinematics.yaml");  // TODO: does not work!
        sen_ikin_ = std::static_pointer_cast<SensorInertialKinematics>(sen_ikin_base);


        // SENSOR + PROCESSOR Imu
        SensorBasePtr sen_imu_base = problem_->installSensor("SensorImu", "SenImu", (Vector7d()<<0,0,0,0,0,0,1).finished(), bodydynamics_root_dir + "/test/sensor_imu.yaml");
        sen_imu_ = std::static_pointer_cast<SensorImu>(sen_imu_base);
        ProcessorBasePtr proc = problem_->installProcessor("ProcessorImu", "Imu pre-integrator", "SenImu", bodydynamics_root_dir + "/test/processor_imu.yaml");
        // auto processor_imu = std::static_pointer_cast<ProcessorImu>(proc);



        ParamsProcessorInertialKinematicsPtr params_ik = std::make_shared<ParamsProcessorInertialKinematics>();
        params_ik->sensor_angvel_name = "SenImu";
        params_ik->std_bp_drift = 0.0000001;
        ProcessorBasePtr proc_ikin_base = problem_->installProcessor("ProcessorInertialKinematics", "PInertialKinematics", sen_ikin_, params_ik);
        // ProcessorBasePtr proc_ikin_base = problem_->installProcessor("ProcessorInertialKinematics", "PInertialKinematics", "SenIK", bodydynamics_root_dir + "/test/processor_inertial_kinematics.yaml");
        // auto proc_inkin = std::static_pointer_cast<ProcessorInertialKinematics>(proc_ikin_base);

        // Set origin of the problem
        x_origin_.resize(19);
        x_origin_ << ZERO3, 0,0,0,1, ZERO3, ZERO3, ZERO3, ZERO3;
        VectorXd std_vec = pow(1e-3, 2)*VectorXd::Ones(18); 
        VectorComposite prior(x_origin_,  "POVCDL", {3,4,3,3,3,3});
        VectorComposite prior_std(std_vec, "POVCDL", {3,3,3,3,3,3});
        problem_->setPriorFactor(prior, prior_std, t_);
        KF0_ = problem_->getTrajectory()->getLastFrame();

        // Set origin of processor Imu
        std::static_pointer_cast<ProcessorImu>(proc)->setOrigin(KF0_);

        ////////////////////////////////////////////
        // Add absolute factor on Imu biases to simulate a fix()
        Matrix6d Q_bi = 1e-8 * Matrix6d::Identity();
        CaptureBasePtr capbi0 = CaptureBase::emplace<CaptureBase>(KF0_, "bi0", t_);
        FeatureBasePtr featbi0 = FeatureBase::emplace<FeatureBase>(capbi0, "bi0", ZERO6, Q_bi);
        FactorBasePtr facbi0 = FactorBase::emplace<FactorBlockAbsolute>(featbi0, featbi0, KF0_->getCaptureOf(sen_imu_)->getSensorIntrinsic(),nullptr,false);

    }

    void TearDown() override{}
};

TEST_F(FactorInertialKinematics_2KF, sensor_and_processors_registration)
{
    TimeStamp t0; t0.set(0);
    TimeStamp t1; t1.set(1);
    TimeStamp t2; t2.set(2);
    TimeStamp t3; t3.set(3);

    // PROCESS ONE Imu CAPTURE
    Vector6d acc_gyr_meas;
    acc_gyr_meas << -gravity(), 0,0,0;
    acc_gyr_meas[0] = 1;
    Matrix6d acc_gyr_cov = 1e-3*Matrix6d::Identity();

    // PROCESS ONE INERTIAL KINEMATICS CAPTURE
    Vector3d pBC_meas = Vector3d::Zero();
    Vector3d vBC_meas = Vector3d::Zero();
    Vector3d w_meas =   Vector3d::Zero();
    // momentum parameters
    Matrix3d Iq; Iq.setIdentity();
    Vector3d Lq = Vector3d::Zero();
    Vector9d meas_ikin0; meas_ikin0 << pBC_meas, vBC_meas, w_meas;

    // sen_imu_->getIntrinsic()->fix();

    // T0
    CaptureImuPtr CImu0 = std::make_shared<CaptureImu>(t0, sen_imu_, acc_gyr_meas, acc_gyr_cov);
    sen_imu_->process(CImu0);
    auto CIKin0 = std::make_shared<CaptureInertialKinematics>(t0, sen_ikin_, meas_ikin0, Iq, Lq);
    sen_ikin_->process(CIKin0);

    CIKin0->getSensorIntrinsic()->fix();

    // T1
    CaptureImuPtr CImu1 = std::make_shared<CaptureImu>(t1, sen_imu_, acc_gyr_meas, acc_gyr_cov);
    sen_imu_->process(CImu1);
    // CImu1->getSensorIntrinsic()->fix();

    auto CIKin1 = std::make_shared<CaptureInertialKinematics>(t1, sen_ikin_, meas_ikin0, Iq, Lq);
    sen_ikin_->process(CIKin1);
    CIKin1->getSensorIntrinsic()->fix();


    // T2
    CaptureImuPtr CImu2 = std::make_shared<CaptureImu>(t2, sen_imu_, acc_gyr_meas, acc_gyr_cov);
    // CImu2->getSensorIntrinsic()->fix();
    sen_imu_->process(CImu2);

    auto CIKin2 = std::make_shared<CaptureInertialKinematics>(t2, sen_ikin_, meas_ikin0, Iq, Lq);
    sen_ikin_->process(CIKin2);
    CIKin2->getSensorIntrinsic()->fix();

    // T3
    CaptureImuPtr CImu3 = std::make_shared<CaptureImu>(t3, sen_imu_, acc_gyr_meas, acc_gyr_cov);
    // CImu3->getSensorIntrinsic()->fix();
    sen_imu_->process(CImu3);

    CaptureInertialKinematicsPtr CIKin3 = std::make_shared<CaptureInertialKinematics>(t3, sen_ikin_, meas_ikin0, Iq, Lq);
    sen_ikin_->process(CIKin3);
    CIKin3->getSensorIntrinsic()->fix();

    KF0_->perturb();

    std::string report = solver_->solve(SolverManager::ReportVerbosity::BRIEF); // 0: nothing, 1: BriefReport, 2: FullReport;
    std::cout << report << std::endl;
    problem_->print(4,true,true,true);

    Vector3d pdiff; pdiff << 4.5, 0, 0;
    Vector3d vdiff; vdiff << 3, 0, 0;

    FrameBasePtr KF1 = problem_->getTrajectory()->getLastFrame();


    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('P')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('V')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('C')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('D')->getState(), ZERO3, 1e-6);
    ASSERT_MATRIX_APPROX(KF0_->getStateBlock('L')->getState(), ZERO3, 1e-6);

    ASSERT_MATRIX_APPROX(KF1->getStateBlock('P')->getState(), pdiff, 1e-6);
    ASSERT_MATRIX_APPROX(KF1->getStateBlock('V')->getState(), vdiff, 1e-6);
    ASSERT_MATRIX_APPROX(KF1->getStateBlock('C')->getState(), pdiff, 1e-6);
    ASSERT_MATRIX_APPROX(KF1->getStateBlock('D')->getState(), vdiff, 1e-6);
    ASSERT_MATRIX_APPROX(KF1->getStateBlock('L')->getState(), ZERO3, 1e-6);
}



int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);

  return RUN_ALL_TESTS();
}
